/***************************************************************************************************
*                           Sample for the usage of class Episodic_TopoART                         *
****************************************************************************************************
*                             Created by Marko Tscherepanow, 27 May 2016                           *
***************************************************************************************************/

// Compile: csc /optimize /r:LibTopoART.dll /r:System.Drawing.dll Episodic_TopoART_sample2.cs
// Run: mono Episodic_TopoART_sample2.exe
// Using mono64 rather than mono may yield a considerable performance gain on suitable systems.

using System;
using System.IO;
using System.Drawing;
using LibTopoART;

namespace LibTopoART_samples
{
	/// <summary>
	/// Episodic clustering sample using real-world video data.
	/// <para>
	/// Like in Section 4.2 of "Marko Tscherepanow, Sina Kühnel, and Sören Riechers (2012). Episodic
	/// Clustering of Data Streams Using a Topology-Learning Neural Network. In Proceedings of the European 
	/// Conference on Artificial Intelligence (ECAI), Workshop on Active and Incremental Learning (AIL), 
	/// pp. 24-29. Montpellier, France.", an Episodic TopoART network is trained with real-world video data. 
	/// Each image has a size of 64x36 pixels. As each pixel comprises 3 color channels (RGB), the input 
	/// dimension equals 6912. After finishing training, recall is performed for a single input stimulus.
	/// </para>
	/// <para>The recall results can be visualised using the R script <c>ShowEpisodicTopoARTRecallResults.R</c> 
	/// provided in the subfolder <c>R</c>.
	/// </para>
	/// </summary>
	class Episodic_TopoART_sample2
	{
		static void Main() {
			
			// Dataset (containing training and test images)
			const string DatasetPath	=	"data/AIL12-like_video_dataset/";
			// Destination directory for trained networks
			const string NetworkPath	=	"results/networks/";
			const long SampleNumber		=	59523;

			// Recall stimulus from the test set (only odd images allowed)
			const string RecallStimulus	=	"image_54971.jpg";
			// Destination directory for recall results
			const string RecallPath		=	"results/recall/AIL12-like_video_dataset_recall_results/";
			// Recall control parameters
			const long MaxInterEpisodeRecallSteps			=	15;
			const long MaxIntraEpisodeRecallSteps			=	50;
			const decimal MinInterEpisodeRecallActivation	=	0.5m;

			// Set to true to omit training and use a saved network file
			bool UseSavedNetwork = false;

			decimal[] InputArray;
			long i, j;
			DateTime start,end;
			long width, height;
			TimeSpan time;
			Episodic_TopoART eta = null;
			string next_image;

			// Start time measuring
			start	=	DateTime.Now;

			// Load dataset and train a new network
			if(!UseSavedNetwork) {

				// Get image size and first image
				i = 0;
				if(LoadImage(DatasetPath + "train/image_" + i.ToString("D5") + ".jpg", out width, out height, out InputArray)) {
					eta				= 	new Episodic_TopoART(InputArray.Length, 2, 0.7m, 400);
					eta.Phi			=	5;
					eta.Tau			=	200;
					eta.Beta_sbm	=	0.25m;

					do {
						if((i % 1000) == 0) Console.Write(".");
						eta.Learn(InputArray);
						i += 2;
					} while((i < SampleNumber) && LoadImage(DatasetPath + "train/image_" + i.ToString("D5") + ".jpg", out width, out height, out InputArray));

					Console.WriteLine(" ");

					// Save network in human readable form
					eta.SaveText(NetworkPath + "Episodic_TopoART_AIL12-like_video_dataset.txt");

					// Save network in binary form
					eta.Save(NetworkPath + "Episodic_TopoART_AIL12-like_video_dataset.eta");
				}
			}
			// Load an existing network
			else {
				eta = new Episodic_TopoART(NetworkPath + "Episodic_TopoART_AIL12-like_video_dataset.eta");
			}

			// Perform recall
			if((eta != null) && LoadImage(DatasetPath + "test/" + RecallStimulus, out width, out height, out InputArray)) {
				decimal[] RecallResult;
				decimal activation;

				// Save recall stimulus
				SaveImage(RecallPath + "stimulus.jpg", width, height, InputArray);

				eta.BeginRecall(InputArray);
				eta.InterEpisodeRecallStep(out RecallResult, out activation);

				// Inter-episode recall loop
				for(i = 1; (RecallResult != null) && (activation > MinInterEpisodeRecallActivation) && (i <= MaxInterEpisodeRecallSteps); ++i) {
					SaveImage(RecallPath + "inter_image_" + i.ToString("D3") + ".jpg", width, height, RecallResult);

					// Intra-episode recall loop
					for(j = 1; (RecallResult != null) && (j <= MaxIntraEpisodeRecallSteps); ++j) {
						eta.IntraEpisodeRecallStep(out RecallResult);
						if(RecallResult == null) 
							break;
						SaveImage(RecallPath + "intra_image_" + i.ToString("D3") + "_" + j.ToString("D3") + ".jpg", width, height, RecallResult);
					}

					// Signify stopped intra-episode recall
					next_image = RecallPath + "intra_image_" + i.ToString("D3") + "_" + j.ToString("D3") + ".jpg";
					if(File.Exists(next_image))
						File.Delete(next_image);

					eta.InterEpisodeRecallStep(out RecallResult, out activation);
				}

				// Signify stopped inter-episode recall
				next_image = RecallPath + "inter_image_" + i.ToString("D3") + ".jpg";
				if(File.Exists(next_image))
					File.Delete(next_image);

				eta.EndRecall();
			}

			// Stop time measuring
			end		=	DateTime.Now;

			// Output the required time
			time	=	end - start;
			Console.WriteLine("time for computation: " + time);
		}

		private static bool LoadImage(string path, out long width, out long height, out decimal[] ImageArray) {
			Bitmap image;
			long ImageSize;
			long PixelNumber;
			bool success = false;
			long j;

			width		=	0;
			height		=	0;

			ImageArray	=	null;

			image		=	new Bitmap(path);

			if(image.PixelFormat != System.Drawing.Imaging.PixelFormat.Format24bppRgb) 
				Console.WriteLine("incompatible pixel format");
			else {
				width	=	image.Width;
				height	=	image.Height;

				PixelNumber		= 	width * height;
				ImageSize		=	PixelNumber * 3;
				ImageArray		=	new decimal[ImageSize];

				for(j = 0; j < PixelNumber; ++j) {
					Color rgb = image.GetPixel((int)(j % width), (int)(j / width));
					ImageArray[j * 3]		=	(decimal)rgb.R / 255.0m;
					ImageArray[j * 3 + 1]	=	(decimal)rgb.G / 255.0m;
					ImageArray[j * 3 + 2]	=	(decimal)rgb.B / 255.0m;
				}

				success = true;
			}

			return success;
		}

		private static void SaveImage(string path, long width, long height, decimal[] ImageArray) {
			long PixelNumber;
			long j;

			Bitmap image = new Bitmap((int)width, (int)height, System.Drawing.Imaging.PixelFormat.Format24bppRgb);

			PixelNumber = ImageArray.LongLength / 3;

			for(j = 0; j < PixelNumber; ++j) {
				Color rgb = Color.FromArgb((int)(ImageArray[j * 3] * 255.0m), (int)(ImageArray[j * 3 + 1] * 255.0m), (int)(ImageArray[j * 3 + 2] * 255.0m));
				image.SetPixel((int)(j % width), (int)(j / width), rgb);
			}

			image.Save(path, System.Drawing.Imaging.ImageFormat.Jpeg);
		}
	}
}
